/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Chroma widget set
 * (C) 2005 M. Derezynski <internalerror@gmail.com>
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>

#include <libchroma/chroma-list.h>
#include <libchroma/chroma-list-column.h>
#include <libchroma/chroma-list-selection.h>
#include <libchroma/chroma-vscrollbar.h>

#define CHROMA_LIST_SELECTION_ALPHA 0.4

static void chroma_list_class_init    (ChromaListClass *klass);
static void chroma_list_init          (ChromaList      *widget_list);
static void chroma_list_realize       (GtkWidget       *widget);
static void chroma_list_size_allocate (GtkWidget       *widget,
				       GtkAllocation   *allocation);
static void chroma_list_send_configure (ChromaList     *widget_list);

static gboolean
chroma_list_drag_leave (ChromaList *self, GdkDragContext *drag_context, guint time_, gpointer user_data);
 
static gboolean
chroma_list_drag_motion (ChromaList *self, GdkDragContext *drag_context, gint x, gint y, guint time, gpointer user_data); 

static gboolean
chroma_list_button_press (ChromaList *self, GdkEventButton *event, gpointer user_data); 

static gboolean
chroma_list_button_release (ChromaList *self, GdkEventButton *event, gpointer user_data);

static gboolean
chroma_list_key_press(ChromaList *self, GdkEventKey *event, gpointer user_data);
 
static gboolean
chroma_list_motion (ChromaList *self, GdkEventMotion *event, gpointer user_data); 

G_LOCK_DEFINE(LIST_WINDOW);

typedef enum {
    CHROMA_LIST_SIGNAL_ROW_ACTIVATED,
    CHROMA_LIST_SIGNAL_CURSOR_POSITION,

    CHROMA_LIST_N_SIGNALS
} ChromaListSignals;
static guint chroma_list_signals[CHROMA_LIST_N_SIGNALS] = {0,0};

typedef enum {
    CHROMA_LIST_PROP_OFFSET = 1,
    CHROMA_LIST_PROP_N_ITEMS,
    CHROMA_LIST_PROP_ROW_HEIGHT,
    CHROMA_LIST_PROP_RESERVED_SPACE,

    CHROMA_LIST_N_PROPERTIES
} ChromaListProperties;

enum {
    SIGNAL_ROW_CHANGED,
    SIGNAL_ROW_DELETED,
    SIGNAL_ROW_INSERTED,
    SIGNAL_ROWS_REORDERED,
    
    N_MODEL_SIGNALS
};

enum {
    SIGNAL_POSITION,
    
    N_ADJUSTMENT_SIGNALS
};

struct _ChromaListPrivate {
    gboolean dispose_has_run;

    GtkTreeModel     *model;
    gulong	      model_signals[N_MODEL_SIGNALS];

    ChromaVScrollbar *adjustment;
    gulong	      adjustment_signals[N_ADJUSTMENT_SIGNALS];

    gint	 *dirty_rows;
    gboolean	  dirty;

    ChromaListColumn *number_column;

    GPtrArray	 *columns;
    gint	 *column_widths;
    gdouble	 *column_xalign;
    gdouble	 *column_yalign;
    gint	  n_columns;

    gint	  row_height;
    gint	  offset;

    gboolean	  active;
    gint	  grabbed_position;

    /* Parameters */
    GtkSelectionMode	      selection_mode;
    ChromaListDropMode	      drop_mode;
    ChromaListAlignMode	      align;
    gboolean		      use_numbers;
    gboolean		      change_on_select;
    gint		      reserved_space;

    gboolean drag;
    gint     drag_row;

    gboolean changed;

    /* Drawing stuff */
    GdkPixmap		     *backingstore;
    PangoFontDescription     *font_description;
    PangoFontMetrics	     *metrics;
    GdkColor		     *foreground;
    GdkColor		     *background;
};

typedef struct {
    ChromaList *self;
    gchar *name;
    gint idx;
} SelectNameTuple;

/* Rendering Fcts Prototypes */
static void
process_cells (ChromaList *self);

static void
backingstore_recreate(ChromaList * self);

static void
backingstore_blank (ChromaList *self);

static void
scroll_up (ChromaList *self);

static void
scroll_down (ChromaList *self);

static void
page_up (ChromaList *self);

static void
page_down (ChromaList *self);

// FIXME: Move out of here
static guint
gint_count_digits (gint n)
{
    guint count = 0;

    n = ABS (n);
    do {
	count++;
	n /= 10;
    } while (n > 0);

    return count;
}

static void
initialize_metrics (ChromaList *self)
{
    self->private->metrics = pango_context_get_metrics (gtk_widget_get_pango_context (GTK_WIDGET (self)),
				        self->private->font_description, 
				        pango_context_get_language(gtk_widget_get_pango_context(GTK_WIDGET (self))));

    self->private->row_height = (pango_font_metrics_get_ascent  (self->private->metrics)/PANGO_SCALE) + 
				(pango_font_metrics_get_descent (self->private->metrics)/PANGO_SCALE) + 3;
}

void
chroma_list_set_selection_mode(ChromaList * self,
                               GtkSelectionMode selection_mode)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    if ((selection_mode != GTK_SELECTION_MULTIPLE) &&
        (selection_mode != GTK_SELECTION_SINGLE))
        return;

    self->private->selection_mode = selection_mode;

}

void
chroma_list_set_column_cell_data_func (ChromaList *self, gint column, ChromaListColumnCellDataFunc cell_data_func)
{
    ChromaListColumn *list_column;

    g_return_if_fail (CHROMA_IS_LIST(self));

    list_column = g_ptr_array_index (self->private->columns, column);
    chroma_list_column_set_cell_data_func (list_column, cell_data_func); 
}

void
chroma_list_set_column_width (ChromaList *self, gint column, gint width)
{
    g_return_if_fail (CHROMA_IS_LIST(self));

    self->private->column_widths[column] = width; 
}

const ChromaListColumn*
chroma_list_get_column (ChromaList *self, gint _idx)
{
    g_return_val_if_fail (CHROMA_IS_LIST(self), NULL);

    return g_ptr_array_index (self->private->columns, _idx);
}

void
chroma_list_set_column_xalign (ChromaList *self, gint _idx, gdouble xalign)
{
    g_return_if_fail (CHROMA_IS_LIST(self));

    self->private->column_xalign[_idx] = xalign;
}

void
chroma_list_set_column_yalign (ChromaList *self, gint _idx, gdouble yalign)
{
    g_return_if_fail (CHROMA_IS_LIST(self));

    self->private->column_yalign[_idx] = yalign;
}

/* k: position index inside current window viewport, iter: iter to render row from */
static void
render_row (ChromaList *self, gint k, GtkTreeIter *iter)
{
    ChromaListColumn *column;
    GdkPixmap *cell;
    GValue value = {0,};
    gint w, h, n, render_index_h = self->private->reserved_space;

    g_return_if_fail (CHROMA_IS_LIST(self));

    if (self->private->use_numbers)
      {
	g_value_init (&value, G_TYPE_INT);  
	g_value_set_int (&value, k+1+self->private->offset);

	chroma_list_column_cell_render (self->private->number_column,
					GTK_WIDGET(self),
					self->private->reserved_space - (pango_font_metrics_get_approximate_digit_width(self->private->metrics)/PANGO_SCALE),	
					&value);	

	g_object_get (self->private->number_column, "cell_width",  &w, NULL);
	g_object_get (self->private->number_column, "cell_height", &h, NULL); // FIXME: Cell height should be not variant, it must be determined by the list itself
	g_object_get (self->private->number_column, "cell", &cell, NULL);	

	gdk_draw_drawable               (GTK_WIDGET(self)->window,
                                         GTK_WIDGET(self)->style->fg_gc[GTK_STATE_NORMAL], 
                                         cell, 
                                         0, 0, // src 
					 0, k*self->private->row_height, // dest
					 w, h); 
	g_value_unset (&value);
      }

    /* First let the columns render the TreeModel data */
    for (n = 0; n < self->private->n_columns; n++) 
      {
	gint cell_set_width;

	gtk_tree_model_get_value        (self->private->model,
                                         iter, 
                                         n, 
                                         &value); 

	column = g_ptr_array_index (self->private->columns, n);

	cell_set_width = self->private->column_widths[n];
	cell_set_width = (cell_set_width == 0) ? GTK_WIDGET(self)->allocation.width / self->private->n_columns : cell_set_width;

	g_object_set (G_OBJECT(column), "xalign", self->private->column_xalign[n], NULL);
	g_object_set (G_OBJECT(column), "yalign", self->private->column_yalign[n], NULL);
	g_object_set (G_OBJECT(column), "foreground", &(GTK_WIDGET(self)->style->fg[GTK_STATE_NORMAL]), NULL);
	g_object_set (G_OBJECT(column), "background", &(GTK_WIDGET(self)->style->bg[GTK_STATE_NORMAL]), NULL);
	g_object_set (G_OBJECT(column), "row_height", self->private->row_height, NULL);

	chroma_list_column_cell_render (column,
					GTK_WIDGET(self),
					cell_set_width, // space allowed to render within	
					&value);	

	g_value_unset (&value);

	g_object_get (column, "cell_width",  &w, NULL);
	g_object_get (column, "cell_height", &h, NULL); // FIXME: Cell height should be not variant, it must be determined by the list itself

	g_object_get (column, "cell", &cell, NULL);	

	gdk_draw_drawable               (GTK_WIDGET(self)->window,
                                         GTK_WIDGET(self)->style->fg_gc[GTK_STATE_NORMAL], 
                                         cell, 
                                         0, 0, // src 
					 render_index_h, k*self->private->row_height, // dest
					 w, h); 

	render_index_h += w; 
      }
}

static void
backingstore_blank (ChromaList *self)
{
	gint w,h;

	gdk_drawable_get_size (self->private->backingstore, &w, &h);
	gdk_draw_rectangle (self->private->backingstore,
			GTK_WIDGET(self)->style->bg_gc[GTK_STATE_NORMAL],	
			TRUE,
			0, 0,
			w,
			h);
}

static void
process_cells (ChromaList *self)
{
    GtkTreePath *path;
    GtkTreeIter iter;
    gint n, max;

    if (!GTK_WIDGET_REALIZED(GTK_WIDGET(self)))
	return;

    if (self->private->dirty_rows)
      {
	path = gtk_tree_path_new_from_indices (self->private->dirty_rows[0]+self->private->offset, -1);
        gtk_tree_model_get_iter (self->private->model, &iter, path);
	gtk_tree_path_free (path);
    
        for (n = 0; self->private->dirty_rows[n] != -1; n++)
	  {
	    render_row (self, self->private->dirty_rows[n], &iter);
	    if (!gtk_tree_model_iter_next (self->private->model, &iter)) break;
	  }

	g_free (self->private->dirty_rows);
        self->private->dirty_rows = NULL;
      }
    else
      {
	path = gtk_tree_path_new_from_indices (self->private->offset, -1);
        gtk_tree_model_get_iter (self->private->model, &iter, path);
	gtk_tree_path_free (path);
    
        for (n = 0; n < chroma_list_get_n_item_display(self); n++)
	  {
	    render_row (self, n, &iter);
	    if (!gtk_tree_model_iter_next (self->private->model, &iter)) break;
	  }
      }

    self->private->dirty = FALSE;
}

static void
backingstore_recreate (ChromaList * self)
{
    if (!GTK_WIDGET_REALIZED(GTK_WIDGET(self)))
      return;

    if (self->private->backingstore)
      {
        gdk_drawable_unref(self->private->backingstore);
	self->private->backingstore = NULL;
      }

    if (self->private->model)
      {
	if ((GTK_WIDGET(self)->allocation.width > 0) && (GTK_WIDGET(self)->allocation.height > 0))
          {

	    self->private->backingstore = gdk_pixmap_new(gtk_widget_get_parent (GTK_WIDGET(self))->window,
					   GTK_WIDGET(self)->allocation.width,
					   GTK_WIDGET(self)->allocation.height,
					   gdk_drawable_get_depth (gtk_widget_get_parent(GTK_WIDGET(self))->window));
	    backingstore_blank (self);
	  }
      }
}

static gboolean
_on_configure_event (ChromaList *self, GdkEventConfigure *event)
{

  if (self->private->adjustment && (chroma_list_get_n_item_display(self) != 0) && (chroma_list_get_n_items(self) != 0))
	chroma_vscrollbar_set_grip_size (self->private->adjustment, 
				         (gdouble)chroma_list_get_n_items(self)/(gdouble)chroma_list_get_n_item_display(self));

  backingstore_recreate (self);
  self->private->dirty = TRUE;

  return FALSE;
}

GtkWidget*
chroma_list_new(void)
{
    ChromaList *self;

    self =
        g_object_new(chroma_list_get_type(), NULL);

    return GTK_WIDGET(self);
}

void
chroma_list_invert_selection(ChromaList * self)
{
}

void
chroma_list_select_nth(ChromaList * self, gint _idx, gboolean selected)
{
}

void
chroma_list_select_all(ChromaList * self)
{
}

void
chroma_list_unselect_all(ChromaList * self)
{
}

static gboolean
chroma_list_button_release(ChromaList * self, GdkEventButton * event, gpointer user_data)
{
    self->private->active = FALSE;
    self->private->grabbed_position = -1;

    return FALSE;
}

static gboolean
chroma_list_drag_leave(ChromaList * self, GdkDragContext *context, guint time_, gpointer user_data)
{
    self->private->drag = FALSE;

    return FALSE;
}

gint
chroma_list_get_drop_position(ChromaList * self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    return self->private->drag_row + self->private->offset;
}

static gboolean
chroma_list_drag_motion (ChromaList *self, GdkDragContext *drag_context, gint x, gint y, guint time, gpointer user_data)
{
    gint control_y;
    gint position;

    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    self->private->drag = TRUE;
    control_y = y;

    if (!self->private->model)
        return FALSE;

    position = control_y / self->private->row_height;

    return FALSE;
}

static gboolean
chroma_list_button_press (ChromaList * self, GdkEventButton * event, gpointer user_data)
{
    GList *itemarray;
    gpointer list_item;
    gint control_y;
    gint position;

    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    control_y = event->y;
    if (!self->private->model)
        return FALSE;

    return FALSE;
}

GList*
chroma_list_get_selected_paths (ChromaList * self)
{
    GList *paths = NULL;
    gpointer list_item;
    gboolean selected = FALSE;
    gint n;

    g_return_val_if_fail(CHROMA_IS_LIST(self), NULL);

    if (!self->private->model)
        return NULL;

    return paths;
}

static gboolean
chroma_list_motion (ChromaList *self, GdkEventMotion *event, gpointer user_data)
{
    gint control_y, position;

    g_return_val_if_fail(CHROMA_IS_LIST(self), FALSE);

    if (self->private->selection_mode == GTK_SELECTION_SINGLE) return FALSE;

    if (!self->private->model)
        return FALSE;

    if (!self->private->active)
        return FALSE;

    control_y = event->y;

    position = control_y / self->private->row_height;
    position += self->private->offset;

    if (self->private->grabbed_position == position)
        return FALSE;

    return FALSE;
}

static gboolean
chroma_list_expose (ChromaList *self, GdkEventExpose *event) 
{
    gint x1_line = 0, y1_line = 0,
         x2_line = 0, y2_line = 0,
         left_padding;
    double r, g, b, a;
    gchar *aux0;
    cairo_t *cr = gdk_cairo_create (GTK_WIDGET(self)->window);

    g_return_val_if_fail(CHROMA_IS_LIST(self), FALSE);

    if (!self->private->model)
	return FALSE;

    if (self->private->backingstore)
      {
        gdk_draw_drawable(GTK_WIDGET(self)->window, 
			  GTK_WIDGET(self)->style->fg_gc[GTK_STATE_NORMAL],
                          self->private->backingstore, 
		          0, 0,
                          0, 0,
                          -1, -1);
      }
    else
      {
        backingstore_recreate (self);
      }

    if (self->private->dirty)
      {
	process_cells (self);
	gdk_draw_drawable (self->private->backingstore,
		       GTK_WIDGET(self)->style->fg_gc[GTK_STATE_NORMAL],
		       GTK_WIDGET(self)->window, 
		       0, 0,
		       0, 0, 
		       -1, -1); 
      }

    if (self->private->use_numbers)
      {
	GdkGCValues values;
	GdkColor color;
#if 0
	left_padding = self->private->font_information->font_width_approx_digits * strlen(aux0) +
		       self->private->font_information->font_width_approx_digits;
#endif

        x1_line =
	    ((gint_count_digits (gtk_tree_model_iter_n_children(CHROMA_LIST(self)->private->model, NULL))+1) *
	    (pango_font_metrics_get_approximate_digit_width(CHROMA_LIST(self)->private->metrics)/PANGO_SCALE)) -
	    ((pango_font_metrics_get_approximate_digit_width(CHROMA_LIST(self)->private->metrics)/PANGO_SCALE)/2.5) + 4;

#if 0
	if (pango_font_description_get_style(self->private->font_information->font_description) != PANGO_STYLE_NORMAL)
	    x1_line += self->private->font_information->font_width_approx_digits;
#endif

        y1_line = 0; 
        x2_line = x1_line;
        y2_line = GTK_WIDGET(self)->allocation.height; 

#if 0
	gdk_gc_get_values (GTK_WIDGET(self)->style->fg_gc[GTK_STATE_NORMAL], &values);
	gdk_color_to_rgba (&(values.foreground), &r, &g, &b, &a);
#endif
	gdk_color_to_rgba (&(GTK_WIDGET(self)->style->fg[GTK_STATE_NORMAL]), &r, &g, &b, &a);

	cairo_save (cr);
	cairo_set_antialias (cr, CAIRO_ANTIALIAS_NONE);
	cairo_set_line_width(cr, 1.0);
	cairo_move_to (cr, x1_line, y1_line);
	cairo_line_to (cr, x1_line, y2_line);
	cairo_set_source_rgba (cr, r, g, b, 0.6);
	cairo_stroke (cr);
	cairo_restore (cr);
      }

#if 0
    else
	left_padding = 0;

    gdk_color_to_rgba(bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT), &r, &g, &b, &a);

    if ((self->private->drag) &&
        (self->private->drop_mode == CHROMA_LIST_DROP_ROWS)) {
        gint lx, ly, lw;

	lx = x1_line + self->private->font_information->font_width_approx_digits;
        ly = (self->private->row_height * self->private->drag_row) + 1;
	lw = GTK_WIDGET(self)->allocation.width - x1_line + self->private->font_information->font_width_approx_digits;

        cairo_set_line_width(cr, 1);
        cairo_move_to(cr, lx, ly);
        cairo_line_to(cr, lx + lw, ly);
        cairo_set_source_rgba(cr, r, g, b, 1.0);
	cairo_set_line_cap (cr, CAIRO_LINE_CAP_ROUND);
	cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
        cairo_stroke(cr);
      }
    else if ((self->private->drag) &&
             (self->private->drop_mode == CHROMA_LIST_DROP_WHOLE))
      {
        gint lx, ly, lw, lh;
        lx = 1; 
        ly = 1; 
        lw = GTK_WIDGET(self)->allocation.width-2; 
        lh = GTK_WIDGET(self)->allocation.height-2; 
        bmp_cairo_draw_rounded_rectangle(cr, lx, ly, lw, lh, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT), 1.0, 1.5, FALSE, 0.35);
      }
#endif

    cairo_destroy (cr);
    return FALSE;
}

#if 0
void
chroma_list_set_font_information(ChromaList * self,
                                 BmpFontInformation * font_information)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->font_information = font_information;

    self->private->row_height =
        self->private->font_information->font_height + 3;

    gtk_widget_queue_draw (GTK_WIDGET(self));
}
#endif

void
chroma_list_set_use_numbers(ChromaList * self, gboolean use_numbers)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->use_numbers = use_numbers;

    gtk_widget_queue_draw (GTK_WIDGET(self));
}

gint
chroma_list_get_n_item_display(ChromaList * self)
{
    gint n_items;

    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (!self->private->row_height) return 0;

    n_items = GTK_WIDGET(self)->allocation.height / self->private->row_height;

    return n_items;
}

gint
chroma_list_get_n_items (ChromaList * self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (!self->private->model)
        return 0;

    return gtk_tree_model_iter_n_children (self->private->model, NULL); 
}

const GtkTreeModel*
chroma_list_get_model (ChromaList * self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), NULL);

    return self->private->model;
}


void
_on_row_changed      (GtkTreeModel *treemodel,
                                            GtkTreePath *arg1,
                                            GtkTreeIter *arg2,
                                            gpointer user_data)
{
}

void
_on_row_deleted      (GtkTreeModel *treemodel,
                                            GtkTreePath *arg1,
                                            gpointer user_data)
{
}

void
_on_row_inserted      (GtkTreeModel *treemodel,
                                            GtkTreePath *arg1,
                                            GtkTreeIter *arg2,
                                            gpointer user_data)
{
}

void
_on_rows_reordered      (GtkTreeModel *treemodel,
		         GtkTreePath *arg1,
			 GtkTreeIter *arg2,
                         gpointer arg3,
                         gpointer user_data)
{
}

void
_on_adjustment_position (ChromaVScrollbar *adjustment,
			 gdouble position,
			 gpointer user_data)
{
    gint offset = 0, n_items = 0, n_item_display = 0;

    ChromaList *list = CHROMA_LIST (user_data);

    n_items =
      chroma_list_get_n_items (list);
    n_item_display =
      chroma_list_get_n_item_display (list);

    if (n_items < n_item_display) return;

    offset = (n_items - n_item_display) * position;

    if (offset >= 0)
      g_object_set (G_OBJECT(list), "offset", offset, NULL);
}

void
chroma_list_set_model (ChromaList * self, GtkTreeModel *tree_model)
{
    ChromaListColumn *column;
    gint n;

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (self->private->model)
      {
	for (n = 0; n < N_MODEL_SIGNALS; n++)
	    g_signal_handler_disconnect (self->private->model, self->private->model_signals[n]); 
	g_object_unref (self->private->model);
      }

    self->private->model = tree_model;

    self->private->model_signals[SIGNAL_ROW_CHANGED] =
	    g_signal_connect_data           ((gpointer)self->private->model,
                                             "row-changed", 
                                             G_CALLBACK(_on_row_changed), 
                                             self, 
                                             NULL, 
                                             0); 

    self->private->model_signals[SIGNAL_ROW_DELETED] =
	    g_signal_connect_data           ((gpointer)self->private->model,
                                             "row-deleted", 
                                             G_CALLBACK(_on_row_changed), 
                                             self, 
                                             NULL, 
                                             0); 

    self->private->model_signals[SIGNAL_ROW_INSERTED] =
	    g_signal_connect_data           ((gpointer)self->private->model,
                                             "row-inserted", 
                                             G_CALLBACK(_on_row_changed), 
                                             self, 
                                             NULL, 
                                             0); 

    self->private->model_signals[SIGNAL_ROWS_REORDERED] =
	    g_signal_connect_data           ((gpointer)self->private->model,
                                             "rows-reordered", 
                                             G_CALLBACK(_on_row_changed), 
                                             self, 
                                             NULL, 
                                             0); 

    if (self->private->columns && (self->private->n_columns > 0))
      {
	for (n = 0; n < self->private->n_columns; n++) 
	  {
	    g_object_unref (g_ptr_array_index (self->private->columns, n));
	    g_ptr_array_remove_fast (self->private->columns, 0);
	  }
	g_ptr_array_free (self->private->columns, FALSE);

	g_free (self->private->column_widths);
	g_free (self->private->column_xalign);
	g_free (self->private->column_yalign);
      }

    self->private->columns = g_ptr_array_new();
    self->private->n_columns = gtk_tree_model_get_n_columns (self->private->model); 

    self->private->column_widths = g_new0(gint, self->private->n_columns); 

    self->private->column_xalign = g_new0(gdouble, self->private->n_columns); 
    self->private->column_yalign = g_new0(gdouble, self->private->n_columns); 

    for (n = 0; n < gtk_tree_model_get_n_columns (self->private->model); n++) 
      {
	column = chroma_list_column_new();		

	g_object_set (column, "font", self->private->font_description, NULL);
	g_object_set (column, "foreground", self->private->foreground, NULL); 
	g_object_set (column, "background", self->private->background, NULL); 

	g_ptr_array_add (self->private->columns, column);

	self->private->column_yalign[n] = 0.5;
	self->private->column_xalign[n] = 0.0;
      }

    g_object_set (self, "offset", (gint)0, NULL);
    self->private->reserved_space = (pango_font_metrics_get_approximate_digit_width (self->private->metrics)/PANGO_SCALE)*(gint_count_digits(gtk_tree_model_iter_n_children(self->private->model,NULL))+2);
  
    backingstore_recreate (self);
    gtk_widget_queue_draw (GTK_WIDGET(self));
}

void
chroma_list_set_adjustment (ChromaList *self, ChromaVScrollbar *adjustment)
{
    gint n;

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (self->private->adjustment)
      {
	for (n = 0; n < N_MODEL_SIGNALS; n++)
	    g_signal_handler_disconnect (self->private->adjustment, self->private->adjustment_signals[n]); 
#if 0
	// FIXME: This requires a ref on setting the adjustment
	g_object_unref (self->private->adjustment);
#endif
      }

    self->private->adjustment = adjustment;

    self->private->adjustment_signals[SIGNAL_POSITION] =
	    g_signal_connect_data           ((gpointer)self->private->adjustment,
                                             "position", 
                                             G_CALLBACK(_on_adjustment_position), 
                                             self, 
                                             NULL, 
                                             0); 

    g_object_connect (G_OBJECT(self), "signal::configure-event", G_CALLBACK(_on_configure_event), NULL, NULL);
}

void
chroma_list_set_drop_mode(ChromaList * self, ChromaListDropMode mode)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->drop_mode = mode;
}

void
chroma_list_set_change_on_select (ChromaList *self, gboolean change_on_select)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->change_on_select = change_on_select;
}

gint
chroma_list_get_row_height (ChromaList *self)
{
    if (! CHROMA_IS_LIST(self))
	return 0;

    return self->private->row_height;
}

#if 0
static void
vs_move (ChromaVScrollbar *vs, gint total, gint offset)
{
    gdouble position, ceil_val;
    gint r_offset;

    ceil_val = 0.00;

    if (offset <= 0)
	position = 0;
    else
    {
	gboolean done = FALSE;

	/* FIXME: We adapt dynamically here which sucks */
	while (! done)
	{
		position = (offset/total)/100. + ceil_val;

		/* Reverse calculation */
	       r_offset = (total) * position;

	       if (r_offset == offset)
		    done = TRUE;
	       else
		    ceil_val += 0.0001;
	}

    }

    if (position < 0) position = 0;
    if (position > 1) position = 1;
    g_object_set(G_OBJECT(vs),"position", position, NULL);

}

static void
scroll_down (ChromaList *self)
{
    ChromaVScrollbar *vs;
    gint offset, n_items, n_item_display;

    vs = g_object_get_data (G_OBJECT(self), "adjustment");
    n_items = g_object_get_int (self, "n_items");
    n_item_display = chroma_list_get_n_item_display(self);
    offset = self->private->offset + 1;

    vs_move (vs, n_items-n_item_display, offset);
}


static void
scroll_up (ChromaList *self)
{
    ChromaVScrollbar *vs;
    gint offset, n_items, n_item_display;

    vs = g_object_get_data (G_OBJECT(self), "adjustment");
    n_items = g_object_get_int (self, "n_items");
    n_item_display = chroma_list_get_n_item_display(self);
    offset = self->private->offset - 1;

    vs_move (vs, n_items-n_item_display, offset);
}

static void
page_up (ChromaList *self)
{
    ChromaVScrollbar *vs;
    gint offset, n_items, n_item_display;

    vs = g_object_get_data (G_OBJECT(self), "adjustment");
    n_items = g_object_get_int (self, "n_items");
    n_item_display = chroma_list_get_n_item_display(self);
    offset = self->private->offset - n_item_display;

    vs_move (vs, n_items-n_item_display, offset);
}

static void
page_down (ChromaList *self)
{
    ChromaVScrollbar *vs;
    gint offset, n_items, n_item_display;

    vs = g_object_get_data (G_OBJECT(self), "adjustment");
    n_items = g_object_get_int (self, "n_items");
    n_item_display = chroma_list_get_n_item_display(self);
    offset = self->private->offset + n_item_display;

    vs_move (vs, n_items-n_item_display, offset);
}

static void
on_vscrollbar_position (ChromaVScrollbar *vs, gdouble position, ChromaList *self)
{
	gint n_items, n_item_display, offset;

	n_items =
	    g_object_get_int (self, "n_items");
	n_item_display =
	    chroma_list_get_n_item_display(self);

	if (n_items < n_item_display) return;

	offset = (n_items - n_item_display) * position;

	if (offset >= 0)
	 	g_object_set (self, "offset", offset, NULL);

	self->private->model->adjustment = position;
}

void
chroma_list_set_adjustment (ChromaList *self, ChromaVScrollbar *adjustment)
{
    g_object_set_data (G_OBJECT(self), "adjustment", adjustment);
    g_signal_connect (adjustment, "position", G_CALLBACK(on_vscrollbar_position), self);
}
#endif


/* GObject stuff */
static void
chroma_list_set_property (GObject *object,
                          guint property_id,
                          const GValue * value,
                          GParamSpec * pspec)
{
    ChromaList *self = (ChromaList *) object;

    switch (property_id) {

    case CHROMA_LIST_PROP_OFFSET: {

	    gint offset_old, delta;

	    offset_old =
		self->private->offset;
    
	    if (!self->private->model)
		return;

	    /* Create the delta between the old and new offset */
	    delta =
		g_value_get_int(value) - offset_old;

	    if (delta == 0)
		break;

	    /* There is no area to copy, we have to re-render wholly */
	    if (abs(delta)>chroma_list_get_n_item_display(self))
              {
                self->private->offset = g_value_get_int(value);
		gtk_widget_queue_draw (GTK_WIDGET(self));
		break;
	      }

	    /* If the delta is positive, we need to move a lower part upwards */
	    if (delta > 0)
              {
		    gint start, height, n, n_start, n_end;
		    gpointer item;

		    start =
			delta;
		    height =
			chroma_list_get_n_item_display(self)-start;

		    gdk_draw_drawable (self->private->backingstore,
				       GTK_WIDGET(self)->style->fg_gc[GTK_STATE_NORMAL],
				       self->private->backingstore,
				       0, start * self->private->row_height,
				       0, 0,
				       -1, height * self->private->row_height);

		    self->private->offset = g_value_get_int(value);

		    self->private->dirty_rows = g_new0 (gint,delta+2);
	
		    for (n = 0; n < delta+1; n++)
		      {
			self->private->dirty_rows[n] = chroma_list_get_n_item_display(self) - 1 - delta + n; 
	              }
		    self->private->dirty_rows[delta+1] = -1; 

		    self->private->dirty = TRUE;
	      }
	    else
              {
		    gint start, height, n, n_start, n_end;
		    gpointer item;

		    start =
			abs(delta);
		    height =
			chroma_list_get_n_item_display(self)-start;

		    gdk_draw_drawable (self->private->backingstore,
				       GTK_WIDGET(self)->style->fg_gc[GTK_STATE_NORMAL],
				       self->private->backingstore,
				       0, 0,
				       0, start * self->private->row_height,
				       -1, height * self->private->row_height);

		    self->private->offset = g_value_get_int(value);

		    self->private->dirty_rows = g_new0 (gint,start+2);

		    for (n = 0; n < start+1; n++)
		      {
			self->private->dirty_rows[n] = n; 
	              }
		    self->private->dirty_rows[start+1] = -1; 

		    self->private->dirty = TRUE;
	      }

            gtk_widget_queue_draw (GTK_WIDGET(self));
        }
        break;

    default:
        /* We don't have any other property... */
        g_assert(FALSE);
        break;
    }
}

static void
chroma_list_get_property(GObject * object,
                         guint property_id,
                         GValue * value, GParamSpec * pspec)
{


    switch (property_id) {

    case CHROMA_LIST_PROP_N_ITEMS:
        {
            if (CHROMA_LIST(object)->private->model == NULL)
                g_value_set_int(value, (gint) 0);
            else
                g_value_set_int(value, gtk_tree_model_iter_n_children (CHROMA_LIST(object)->private->model, NULL));
        }
        break;

    case CHROMA_LIST_PROP_RESERVED_SPACE:
        {
            g_value_set_int(value, CHROMA_LIST(object)->private->reserved_space);
        }
        break;

    case CHROMA_LIST_PROP_ROW_HEIGHT:
        {
            g_value_set_int(value, CHROMA_LIST(object)->private->row_height);
        }
        break;

    default:
        /* We don't have any other property... */
        g_assert(FALSE);
        break;
    }
}

static void
chroma_list_dispose(GObject * obj)
{
    ChromaList *self = (ChromaList *) obj;

    if (self->private->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
    /* Make sure dispose does not run twice. */
    self->private->dispose_has_run = TRUE;

    /*
     * In dispose, you are supposed to free all types referenced from this
     * object which might themselves hold a reference to self. Generally,
     * the most simple solution is to unref all members on which you own a
     * reference.
     */
}

static void
chroma_list_finalize(GObject * obj)
{
    ChromaList *self = (ChromaList *) obj;

    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */
    g_free(self->private);
}

static GObject *
chroma_list_constructor(GType type,
                        guint n_construct_properties,
                        GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        /* Invoke parent constructor. */
        ChromaListClass *klass;
        GObjectClass *parent_class;
        klass = CHROMA_LIST_CLASS(g_type_class_peek(CHROMA_TYPE_LIST));
        parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
        obj = parent_class->constructor(type,
                                        n_construct_properties,
                                        construct_properties);
    }

    /* do stuff. */

    return obj;
}

static void
chroma_list_class_init(ChromaListClass *g_class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (g_class);
    GParamSpec *pspec;

    widget_class->realize = chroma_list_realize;
    widget_class->size_allocate = chroma_list_size_allocate;
    widget_class->drag_leave = chroma_list_drag_leave;
    widget_class->drag_motion = chroma_list_drag_motion;
    widget_class->button_press_event = chroma_list_button_press; 
    widget_class->button_release_event = chroma_list_button_release;
#if 0
    widget_class->key_press_event = chroma_list_key_press;
#endif
    widget_class->motion_notify_event = chroma_list_motion;
    widget_class->expose_event = chroma_list_expose;

    gobject_class->set_property = chroma_list_set_property;
    gobject_class->get_property = chroma_list_get_property;
    gobject_class->dispose = chroma_list_dispose;
    gobject_class->finalize = chroma_list_finalize;
    gobject_class->constructor = chroma_list_constructor;

    pspec = g_param_spec_int("reserved_space",
                             "reserved_space",
                             "reserved_space", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class,
                                    CHROMA_LIST_PROP_RESERVED_SPACE, pspec);

    pspec = g_param_spec_int("offset",
                             "offset",
                             "offset", 0, G_MAXINT, 0, G_PARAM_WRITABLE);
    g_object_class_install_property(gobject_class,
                                    CHROMA_LIST_PROP_OFFSET, pspec);

    pspec = g_param_spec_int("n_items",
                             "n_items",
                             "n_items", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class,
                                    CHROMA_LIST_PROP_N_ITEMS, pspec);

    pspec = g_param_spec_int("row_height",
                             "row_height",
                             "row_height", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class,
                                    CHROMA_LIST_PROP_ROW_HEIGHT, pspec);

    chroma_list_signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION] =
        g_signal_new("cursor-position",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, cursor_position),
                     NULL, NULL,
                     g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);

    chroma_list_signals[CHROMA_LIST_SIGNAL_ROW_ACTIVATED] =
        g_signal_new("row_activated",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, row_activated),
                     NULL, NULL,
                     g_cclosure_marshal_VOID__INT,
                     G_TYPE_NONE, 1, G_TYPE_INT);

}

static void
chroma_list_init (ChromaList *self)
{
    self->private = g_new(ChromaListPrivate, 1);
    self->private->dispose_has_run = FALSE;
    self->private->model = NULL;
    self->private->adjustment = NULL;
    self->private->dirty_rows = NULL;
    self->private->dirty = TRUE;
    self->private->use_numbers = TRUE;
    self->private->offset = 0;
    self->private->align = CHROMA_LIST_ALIGN_LEFT;
    self->private->n_columns = 1;
    self->private->selection_mode = GTK_SELECTION_MULTIPLE;
    self->private->change_on_select = FALSE;
    self->private->active = FALSE;
    self->private->grabbed_position = 0;
    self->private->drag = FALSE;
    self->private->drag_row = -1;
    self->private->changed = TRUE;
    self->private->drop_mode = CHROMA_LIST_DROP_ROWS;
    self->private->backingstore = NULL;
    self->private->columns = NULL;
    self->private->column_widths = NULL;
    self->private->n_columns = 0;
    self->private->foreground = g_new0 (GdkColor,1);
    self->private->background = g_new0 (GdkColor,1);
    self->private->number_column = chroma_list_column_new();		
    self->private->row_height = 0; 
    self->private->reserved_space = 0;
    self->private->font_description = pango_font_description_from_string ("Bitstream Vera Sans 10"); 

    gdk_color_parse ("#4d5563", self->private->foreground);
    gdk_color_parse ("#f1f7e9", self->private->background);

    gtk_widget_modify_bg (GTK_WIDGET(self), GTK_STATE_NORMAL, self->private->background);
    gtk_widget_modify_fg (GTK_WIDGET(self), GTK_STATE_NORMAL, self->private->foreground);

    initialize_metrics (self);

    g_object_set (self->private->number_column, "foreground", self->private->foreground, NULL); 
    g_object_set (self->private->number_column, "background", self->private->background, NULL); 
    g_object_set (self->private->number_column, "font", self->private->font_description, NULL);
    g_object_set (self->private->number_column, "row_height", self->private->row_height, NULL);
    g_object_set (G_OBJECT(self->private->number_column), "xalign", 1.0, NULL);
    g_object_set (G_OBJECT(self->private->number_column), "yalign", 0.5, NULL);
}

void
chroma_list_size (ChromaList *widget_list,
		       gint            width,
		       gint            height)
{
  g_return_if_fail (CHROMA_IS_LIST (widget_list));

  GTK_WIDGET (widget_list)->requisition.width = width;
  GTK_WIDGET (widget_list)->requisition.height = height;

  gtk_widget_queue_resize (GTK_WIDGET (widget_list));
}

static void
chroma_list_realize (GtkWidget *widget)
{
  ChromaList *widget_list;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (CHROMA_IS_LIST (widget));

  widget_list = CHROMA_LIST (widget);
  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK | GDK_ALL_EVENTS_MASK;

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, widget_list);

  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  chroma_list_send_configure (CHROMA_LIST (widget));
}

static void
chroma_list_size_allocate (GtkWidget     *widget,
				GtkAllocation *allocation)
{
  g_return_if_fail (CHROMA_IS_LIST (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

      chroma_list_send_configure (CHROMA_LIST (widget));
    }
}

static void
chroma_list_send_configure (ChromaList *self)
{
  GtkWidget *widget;
  GdkEvent *event = gdk_event_new (GDK_CONFIGURE);

  widget = GTK_WIDGET (self);

  event->configure.window = g_object_ref (widget->window);
  event->configure.send_event = TRUE;
  event->configure.x = widget->allocation.x;
  event->configure.y = widget->allocation.y;
  event->configure.width = widget->allocation.width;
  event->configure.height = widget->allocation.height;

  gtk_widget_event (widget, event);
  gdk_event_free (event);
}

static void
chroma_list_selection_init (ChromaListSelectionInterface *iface)
{
    iface->get_selected_paths	 = NULL; 
    iface->iter_get_selected	 = NULL; 
    iface->iter_set_selected	 = NULL; 
}

GType
chroma_list_get_type (void)
{
    static GType list_type = 0;

    if (!list_type)
     {
         static const GTypeInfo list_info = {
             sizeof (ChromaListClass),
             NULL,
             NULL,
             (GClassInitFunc) chroma_list_class_init,
             NULL,
             NULL,
             sizeof (ChromaList),
             0,
             (GInstanceInitFunc) chroma_list_init,
         };

         static const GInterfaceInfo list_selection_info = {
             (GInterfaceInitFunc) chroma_list_selection_init,
             NULL,
             NULL
         };

         list_type =
             g_type_register_static (GTK_TYPE_WIDGET, "ChromaList",
                                     &list_info, 0);

         g_type_add_interface_static (list_type,
                                      CHROMA_TYPE_LIST_SELECTION, 
				      &list_selection_info);

     }

    return list_type;
}
